require 'wavefile'

module STFTSpectrogram
  # Represents wave file
  class AudioFile
    include WaveFile

    attr_reader :window_overlap
    attr_reader :window_size

    SAMPLE_RATE = 441_00

    def initialize(path, wsize = 1, woverlap = 0)
      unless File.file?(path)
        raise IOError, "File '" + path + "' does not exist!"
      end
      @samples = []
      self.window_size = wsize
      self.window_overlap = woverlap
      @window_overlap = @window_size / 2 if @window_size <= @window_overlap
      @window_pos = 0
      read_data(path)
    end

    def read_data(path)
      format = Format.new(:mono, :float, SAMPLE_RATE)
      Reader.new(path, format).each_buffer do |buffer|
        buffer.samples.each { |s| @samples.push(s) }
      end
    end

    # Nearest power of 2 to num
    def nearest_pow2(num)
      return 0 if num <= 0
      2**Math.log2(num).round
    end

    def window_overlap=(overlap)
      @window_overlap = nearest_pow2(overlap * (SAMPLE_RATE / 1000))
    end

    def window_size=(size)
      @window_size = nearest_pow2(size * (SAMPLE_RATE / 1000))
    end

    # Gets window_size of data from audio samples
    # and moves the window pointer
    def next_window
      return [] if end?

      slice = @samples[@window_pos, @window_size]
      slice.fill(0, slice.length..@window_size - 1)
      @window_pos += @window_size - @window_overlap

      slice
    end

    def self.sample_rate
      SAMPLE_RATE
    end

    # Resets the window pointer
    def reset
      @window_pos = 0
    end

    # Gets the time corresponding to the window pointers location
    def current_time
      (@window_pos + @window_size / 2) / (SAMPLE_RATE / 1000)
    end

    # Checks if the window pointer is at the end of the file
    def end?
      @window_pos + @window_size >= @samples.length
    end

    private :read_data, :nearest_pow2
  end
end
