require 'numo/gnuplot'
require 'tempfile'

module STFTSpectrogram
  # Creates plots using gnuplot
  class Plot
    TEMP_SPECTR_IMG_FILE = 'spec'.freeze
    TEMP_SPECTR_DAT_FILE = 'spec.dat'.freeze

    attr_reader :imgfile

    def initialize
      @imgfile = Tempfile.new([TEMP_SPECTR_IMG_FILE, '.png'])
      @imgfile.close
      set_gnuplot_defaults
    end

    def set_gnuplot_defaults
      Numo.gnuplot do
        set title: ''
        set palette: 'rgb 34,35,36'
        set cblabel: 'Magnitude'
        set xtics: 'auto'
        set ytics: 'auto'
        set xlabel: 'Time (s)'
        set ylabel: 'Frequency (kHz)'
      end
    end

    # Sets x axis tics
    def xtics=(str)
      Numo.gnuplot { set xtics: str }
    end

    # Sets y axis tics
    def ytics=(str)
      Numo.gnuplot { set ytics: str }
    end

    def write_temp_data(data)
      datafile = Tempfile.new(TEMP_SPECTR_DAT_FILE)
      datafile.write(data.map { |row| row.join(' ') }.join("\n"))
      datafile.close
      datafile
    end

    def build_term_str(w, h, fnt_size)
      'png font arial ' + fnt_size.to_s + ' size ' + w.to_s + ',' + h.to_s
    end

    # Plots data in a matrix format using a temporary file
    # Data are plotted to 2D with their magnitude being differentiated
    # by color gradient
    def plot_matrix(data, w = 1920, h = 1080, fnt_size = 12)
      tmpfile = write_temp_data(data)
      outfile = @imgfile.path
      termstr = build_term_str(w, h, fnt_size)
      Numo.gnuplot do
        set term: termstr
        set output: outfile
        plot "'" + tmpfile.path + "'", :matrix, w: 'image', t: ''
        unset :output
      end
      tmpfile.unlink
    end

    # Plots data contained int x and y parameters
    def plot(x, y, w = 1920, h = 1080, fnt_size = 12)
      outfile = @imgfile.path
      termstr = build_term_str(w, h, fnt_size)
      Numo.gnuplot do
        set term: termstr
        set output: outfile
        set autoscale: 'xfix'
        set autoscale: 'yfix'
        set autoscale: 'cbfix'
        plot x, y, w: 'lines', t: '', lc_rgb: 'red'
        unset :output
      end
    end

    private :set_gnuplot_defaults
  end
end
