require_relative 'fft.rb'
require_relative '../audio/audio_file.rb'

module STFTSpectrogram
  # Represents a data slic. Performs FFT
  class STFTSlice
    attr_reader :data
    attr_reader :time

    attr_writer :low
    attr_writer :high

    def initialize(data, time = 0)
      @spectrum = {}
      @data = data.product([0]).flatten
      @time = time / 1000.0 # ms to seconds
      @low = 0
      @high = 0
    end

    # Performs FFT on this data window
    def do_fft!
      FFT.do_fft(@data)
      @data, = @data.drop(2).each_slice((@data.size / 2.0).round).to_a
      create_spectrum!
    end

    # Creates a spectrogram from the transformed data - frequencies
    # with their magnitudes
    # Spectrogram is saved in a Hashmap, with frequency as the key
    def create_spectrum!
      # l, _r = @data.drop(2).each_slice((@data.size / 2.0).round).to_a
      @data.each_slice(2).with_index do |(real, img), i|
        freq = (i * (AudioFile.sample_rate / @data.length))
        magnitude = Math.sqrt(real * real + img * img)
        @spectrum[freq] = magnitude
      end
    end

    # Returns strongest frequency with its magnitude
    def strongest_freq
      spectrum.max_by { |_k, v| v }
    end

    # Returns highest frequency in this time window
    def max_freq
      freqs.max
    end

    # Returns frequencies
    def freqs
      ret = @spectrum.keys
      ret.select! { |f| f >= @low } unless @low.zero?
      ret.select! { |f| f <= @high } unless @high.zero?
      ret
    end

    # Returns frequencies with their magnitudes in this time window
    def spectrum
      ret = @spectrum.dup
      ret.select! { |f, _m| f >= @low } unless @low.zero?
      ret.select! { |f, _m| f <= @high } unless @high.zero?
      ret
    end

    private :create_spectrum!
  end
end
